import * as THREE from '../build/three.module.js';
import Stats from './jsm/libs/stats.module.js';

class App {
    constructor() {
        const domWebGL = document.createElement('div');
        document.body.appendChild(domWebGL);
        this._domWebGL = domWebGL;

        const renderer = new THREE.WebGLRenderer({ antialias: true });
        //renderer.outputEncoding = THREE.sRGBEncoding;
        renderer.setPixelRatio(window.devicePixelRatio);
        //renderer.shadowMap.enabled =  true;
        domWebGL.appendChild(renderer.domElement);
        this._renderer = renderer;

        const scene = new THREE.Scene();
        scene.background = new THREE.Color(0x050505);
        scene.fog = new THREE.Fog(0x050505, 2000, 3500);
        this._scene = scene;

        this._setupCamera();
        //this._setupLight();
        this._setupModel();
        //this._setupInteractions();
        this._setupGUI();
        //this._setupControls();
        //this._setupHelper();
        
        window.onresize = this.resize.bind(this);

        this.resize();
        this.render()
    }

    _setupInteractions() {
        //.
    }

    _setupPostprocessing() {
        //.
    }

    _setupModel() {
        const geometry = new THREE.BufferGeometry();
        const positions = [];
        const positions2 = [];
        const colors = [];
        const color = new THREE.Color();
        const n = 1000, n2=n/2;
        const particles = 300000;
        const drawCount = 10000;

        for(let i=0; i<particles; i++) {
            const x = Math.random() * n - n2; 
            const y = Math.random() * n - n2; 
            const z = Math.random() * n - n2;

            positions.push(x, y, z);
            positions2.push(x * 0.3, y*0.3, z*0.3);

            color.setRGB((x/n)+0.5, (y/n)+0.5, (z/n)+0.5);
            colors.push(color.r, color.g, color.b);
        }

        const gl = this._renderer.getContext();
        const pos = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, pos);
        gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(positions), gl.STATIC_DRAW);

        const pos2 = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, pos2);
        gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(positions2), gl.STATIC_DRAW);

        const rgb = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, rgb);
        gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(colors), gl.STATIC_DRAW);

        const posAttr1 = new THREE.GLBufferAttribute(pos, gl.FLOAT, 3, 4, particles);
        const posAttr2 = new THREE.GLBufferAttribute(pos2, gl.FLOAT, 3, 4, particles);
        geometry.setAttribute("position", posAttr1);

        setInterval(() => {
            const attr = geometry.getAttribute("position");
            geometry.setAttribute("position", (attr === posAttr1) ? posAttr2 : posAttr1);
        }, 2000);

        geometry.setAttribute("color", new THREE.GLBufferAttribute(rgb, gl.FLOAT, 3, 4, particles));

        const material = new THREE.PointsMaterial({ size: 15, vertexColors: true });
        const points = new THREE.Points(geometry, material);

        // Choose one:
        // geometry.boundingSphere = ( new THREE.Sphere() ).set( new THREE.Vector3(), Infinity );
        points.frustumCulled = false; // 이 코드가 왜 필요하지?

        this._scene.add(points);

        this._points = points;
        this._particles = particles;
        this._drawCount = drawCount;
    }

    _setupHelper() {
        //.
    }

    _setupGUI() {
        const stats = new Stats();
        this._domWebGL.appendChild(stats.dom);
        this._stats = stats;
    }

    _setupCamera() {
        const camera = new THREE.PerspectiveCamera(27, window.innerWidth / window.innerHeight, 5, 3500);
        camera.position.z = 2750;

        this._camera = camera;
    }

    _setupControls() {
        //.
    }

    _setupLight() {
        //.
    }

    update() {
        const time = Date.now() * 0.001;

        this._drawCount = (Math.max(1, this._drawCount) + Math.floor(5000 * Math.sin(time*5))) % this._particles;
        this._points.geometry.setDrawRange(0, this._drawCount);

        this._points.rotation.x = time * 0.1;
        this._points.rotation.y = time * 0.2;
        
        this._stats.update();
    }

    render() {
        requestAnimationFrame(this.render.bind(this));
        
        this.update();
        this._renderer.render(this._scene, this._camera);
    }

    resize() {
        this._camera.aspect = window.innerWidth / window.innerHeight;
        this._camera.updateProjectionMatrix();
        this._renderer.setSize(window.innerWidth, window.innerHeight);
    }
}

window.onload = function () {
    new App();
}